import pandas as pd
import random
import numpy as np
import os
import time
from math import ceil
from .utils import TimeCounter
from .LoadingData.data_construction import load_scope_data, sample_seqs, data_split
from .vector_construction import mp_get_vec


def remove_temp_files(args, temp_path):
    os.remove(f"{temp_path}_ori.seq")
    os.remove(f"{temp_path}_pos.seq")
    os.remove(f"{temp_path}_neg.seq")
    os.remove(f"{temp_path}_ori.seq.npy")
    os.remove(f"{temp_path}_pos.seq.npy")
    os.remove(f"{temp_path}_neg.seq.npy")
    os.remove(f"generation_done")
    
    for i in range(args.total_rank):
        ori_path = f"{temp_path}_ori.seq_{i}.npy"
        pos_path = f"{temp_path}_pos.seq_{i}.npy"
        neg_path = f"{temp_path}_neg.seq_{i}.npy"
        os.remove(ori_path)
        os.remove(pos_path)
        os.remove(neg_path)


def write_sequences(seqs, save_path):
    with open(save_path, 'w') as f:
        for seq in seqs:
            f.write(f"{seq}\n")


def load_sequences(path):
    with open(path, 'r') as f:
        return [seq[:-1] for seq in f.readlines()]


def generate_training_set(data, path, neg_num, n_per_sf):
    fasta_path = "/sujin/dataset/uniclust30/uniclust30_2018_08_seed.fasta"
    pointer_path = "/sujin/dataset/uniclust30/uniclust30_2018_08_seed_pointer.tsv"
    pointer = pd.read_csv(pointer_path, sep='\t').values
    
    # sampling negative samples
    neg_samples = sample_seqs(fasta_path, pointer, neg_num, 512)
    
    # sampling original and positive samples
    ori_seqs, pos_seqs = [], []
    for records in data.values():
        for i in range(n_per_sf):
            ori, pos = random.sample(records, 2)
            ori_seqs.append(str(ori.seq))
            pos_seqs.append(str(pos.seq))
    
    write_sequences(neg_samples, f"{path}_neg.seq")
    write_sequences(ori_seqs, f"{path}_ori.seq")
    write_sequences(pos_seqs, f"{path}_pos.seq")

    with open(f"generation_done", 'w') as f:
        pass


def calculate_vectors(args, model, temp_path):
    device_ids = list(range(args.local_size))
    seqs = load_sequences(temp_path)
 
    len_per_part = ceil(len(seqs)/args.world_size)
    data_start = len_per_part * args.start_loc
    data_end = data_start + len_per_part * args.local_size
    sub_seqs = seqs[data_start:data_end]
    
    # calculating vectors of negative samples
    save_path = f"{temp_path}_{args.node_rank}.npy"
    mp_get_vec(device_ids, model, sub_seqs, save_path=save_path, batch_size=64)
    
    # wait until all vectors calculated
    while True:
        flag = True
        for i in range(args.total_rank):
            path = f"{temp_path}_{i}.npy"
            if not os.path.exists(path):
                flag = False
                break
        
        if flag:
            break
    
    # master host merges several vectors and others wait
    if args.node_rank == 0:
        vectors = np.empty((len(seqs), model.dim), dtype=np.float32)
        pointer = 0
        for i in range(args.total_rank):
            path = f"{temp_path}_{i}.npy"
            v = np.load(path)
            vectors[pointer: pointer + v.shape[0]] = v
            pointer += v.shape[0]
    
        np.save(f"{temp_path}.npy", vectors)
    
    else:
        while True:
            if os.path.exists(f"{temp_path}.npy"):
                break


def build_training_file(args, temp_path, n_per_sf, neg_per_sample):
    if args.node_rank == 0:
        ori_seqs = load_sequences(f"{temp_path}_ori.seq")
        pos_seqs = load_sequences(f"{temp_path}_pos.seq")
        neg_seqs = load_sequences(f"{temp_path}_neg.seq")
        
        ori_vecs = np.load(f"{temp_path}_ori.seq.npy")
        pos_vecs = np.load(f"{temp_path}_pos.seq.npy")
        neg_vecs = np.load(f"{temp_path}_neg.seq.npy")
        
        # choose the negative sample with nearest distance to the original sample
        # and write results into a tsv format file
        with open(f"{temp_path}.tsv", 'w') as f:
            f.write("ori\tpos\tneg\n")
        
            for i, ori_vec in enumerate(ori_vecs):
                st = int(i / n_per_sf) * n_per_sf
                sub_pos_vecs = pos_vecs[st: st+n_per_sf]
                pos_dist = np.linalg.norm(sub_pos_vecs - ori_vec, ord=2, axis=1)
                max_loc = np.argmax(pos_dist)
                pos_seq = pos_seqs[st+max_loc]

                indices = [random.randint(0, len(neg_seqs) - 1) for _ in range(neg_per_sample)]
                sub_neg_vecs = neg_vecs[indices]
                
                dist = np.linalg.norm(sub_neg_vecs - ori_vec, ord=2, axis=1)
                min_loc = np.argmin(dist)
                if dist[min_loc] != np.inf and pos_dist[max_loc] != np.inf:
                    f.write(f"{ori_seqs[i]}\t{pos_seq}\t{neg_seqs[indices[min_loc]]}\n")
        
        # master host sends a sign indicating the work done
        with open(f"{temp_path}_0_validation", 'w') as f:
            pass
    
    else:
        while True:
            if os.path.exists(f"{temp_path}_0_validation"):
                break


def dynamic_sampling(args, data, model, temp_path, n_per_sf, neg_num, neg_per_sample):
    if not os.path.exists(f"{temp_path}.tsv"):
        # master computer generates training set
        if args.node_rank == 0:
            generate_training_set(data, temp_path, neg_num, n_per_sf)
        
        # others wait until the training set built
        else:
            while True:
                if os.path.exists(f"generation_done"):
                    break
    
        # divide training set into pieces for each host to calculate vectors
        # ori seqs
        calculate_vectors(args, model, f"{temp_path}_ori.seq")
        # pos seqs
        calculate_vectors(args, model, f"{temp_path}_pos.seq")
        # neg_seqs
        calculate_vectors(args, model, f"{temp_path}_neg.seq")
        
        # master host builds training file
        build_training_file(args, temp_path, n_per_sf, neg_per_sample)
        
        if args.node_rank == 0:
            # check if every host breaks the loop
            while True:
                flag = True
                for i in range(1, args.total_rank):
                    path = f"{temp_path}_{i}_validation"
                    if not os.path.exists(path):
                        flag = False
                        break
    
                if flag:
                    break
                
            remove_temp_files(args, temp_path)
            for i in range(args.total_rank):
                os.remove(f"{temp_path}_{i}_validation")
                
        else:
            with open(f"{temp_path}_{args.node_rank}_validation", 'w') as f:
                pass
    


# Copyright (c) Facebook, Inc. and its affiliates.
# 1
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from .version import version as __version__  # noqa

from .data import Alphabet, BatchConverter, FastaBatchedDataset  # noqa
from .model import ProteinBertModel, MSATransformer  # noqa
from . import pretrained  # noqa



import os
import random
import multiprocessing as mp
import pandas as pd
from Bio import SeqIO
from tqdm import tqdm
from utils import setup_seed, progress_bar
from math import ceil


def building(file_list, path):
	print("building training set...")
	with open(path, 'w') as f:
		f.write("ori\tpos\tneg\n")

		for i, file in enumerate(file_list):
			sample_n = 100

			records = list(SeqIO.parse(file, 'fasta'))
			ori = str(records[0].seq)

			# 对正样本进行采样
			pos_list = records if len(records) == 1 else records[1:]
			pos_seqs = []
			for _ in range(sample_n):
				pos_seqs.append(str(random.choice(pos_list).seq).replace('-', '').upper())

			# 对负样本进行采样
			while True:
				neg_files = random.sample(file_list, sample_n)
				if file not in neg_files:
					break

			neg_seqs = []
			for file in neg_files:
				neg_records = list(SeqIO.parse(file, 'fasta'))
				sample = random.choice(neg_records)
				neg_seqs.append(str(sample.seq).replace('-', '').upper())

			for pos, neg in tuple(zip(*[pos_seqs, neg_seqs])):
				f.write(f"{ori}\t{pos}\t{neg}\n")

			progress_bar(i+1, len(file_list), path)


def mp_setup(path, n):
	print("loading data...")
	print(path)
	dirs = os.listdir(path)
	file_list = []
	for dir in tqdm(dirs):
		sub_dir = path + "/" + dir
		for file in os.listdir(sub_dir):
			file_path = sub_dir + '/' + file
			file_list.append(file_path)
	print("finished")

	# 将数据分割为多份
	num = ceil(len(file_list) / n)
	p_list = []
	for i in range(n):
		st = num * i
		ed = st + num
		path = f"../Data/PkU_training_set_{st}_{ed}.tsv"

		p = mp.Process(target=building, args=(file_list[st: ed], path))
		p.start()
		p_list.append(p)

	# 等待所有进程执行完毕
	for p in p_list:
		p.join()

	# 合并数据
	count = 0
	with open("../Data/PkU_training_set.tsv", 'w') as f:
		f.write("ori\tpos\tneg\n")

		for i in range(n):
			st = num * i
			ed = st + num
			path = f"../Data/PkU_training_set_{st}_{ed}.tsv"
			with open(path, 'r') as rf:
				for line in rf.readlines()[1:]:
					f.write(line)

			os.remove(path)


if __name__ == '__main__':
	setup_seed(2021)
	path = "../Data/PkU_training_set.tsv"
	# mp_setup(path, 100)

	data = pd.read_csv(path, sep='\t')
	tiny_data = data.sample(4000)
	print(tiny_data)
	tiny_data.to_csv("../Data/PkU_tiny_training_set.tsv", index=False, sep='\t')


# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import math

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.checkpoint import checkpoint

from .modules import (
    TransformerLayer,
    AxialTransformerLayer,
    LearnedPositionalEmbedding,
    SinusoidalPositionalEmbedding,
    RobertaLMHead,
    ESM1bLayerNorm,
    ContactPredictionHead,
)

from .axial_attention import RowSelfAttention, ColumnSelfAttention


class ProteinBertModel(nn.Module):
    @classmethod
    def add_args(cls, parser):
        parser.add_argument(
            "--num_layers", default=36, type=int, metavar="N", help="number of layers"
        )
        parser.add_argument(
            "--embed_dim", default=1280, type=int, metavar="N", help="embedding dimension"
        )
        parser.add_argument(
            "--logit_bias", action="store_true", help="whether to apply bias to logits"
        )
        parser.add_argument(
            "--ffn_embed_dim",
            default=5120,
            type=int,
            metavar="N",
            help="embedding dimension for FFN",
        )
        parser.add_argument(
            "--attention_heads",
            default=20,
            type=int,
            metavar="N",
            help="number of attention heads",
        )

    def __init__(self, args, alphabet):
        super().__init__()
        self.args = args
        self.alphabet_size = len(alphabet)
        self.padding_idx = alphabet.padding_idx
        self.mask_idx = alphabet.mask_idx
        self.cls_idx = alphabet.cls_idx
        self.eos_idx = alphabet.eos_idx
        self.prepend_bos = alphabet.prepend_bos
        self.append_eos = alphabet.append_eos
        if self.args.arch == 'roberta_large':
            self.model_version = 'ESM-1b'
            self._init_submodules_esm1b()
        else:
            self.model_version = 'ESM-1'
            self._init_submodules_esm1()

    def _init_submodules_common(self):
        self.embed_tokens = nn.Embedding(
            self.alphabet_size, self.args.embed_dim, padding_idx=self.padding_idx
        )
        self.layers = nn.ModuleList(
            [
                TransformerLayer(
                    self.args.embed_dim, self.args.ffn_embed_dim, self.args.attention_heads,
                    add_bias_kv=(self.model_version != 'ESM-1b'),
                    use_esm1b_layer_norm=(self.model_version == 'ESM-1b'),
                )
                for _ in range(self.args.layers)
            ]
        )

        self.contact_head = ContactPredictionHead(
            self.args.layers * self.args.attention_heads,
            self.prepend_bos,
            self.append_eos,
            eos_idx=self.eos_idx,
        )

    def _init_submodules_esm1b(self):
        self._init_submodules_common()
        self.embed_scale = 1
        self.embed_positions = LearnedPositionalEmbedding(self.args.max_positions, self.args.embed_dim, self.padding_idx)
        self.emb_layer_norm_before = ESM1bLayerNorm(self.args.embed_dim)
        self.emb_layer_norm_after = ESM1bLayerNorm(self.args.embed_dim)
        self.lm_head = RobertaLMHead(
            embed_dim=self.args.embed_dim,
            output_dim=self.alphabet_size,
            weight=self.embed_tokens.weight
        )

    def _init_submodules_esm1(self):
        self._init_submodules_common()
        self.embed_scale = math.sqrt(self.args.embed_dim)
        self.embed_positions = SinusoidalPositionalEmbedding(self.args.embed_dim, self.padding_idx)
        self.embed_out = nn.Parameter(
            torch.zeros((self.alphabet_size, self.args.embed_dim))
        )
        self.embed_out_bias = None
        if self.args.final_bias:
            self.embed_out_bias = nn.Parameter(torch.zeros(self.alphabet_size))

    def forward(self, tokens, repr_layers=[], use_checkpoint=False):
        assert tokens.ndim == 2
        padding_mask = tokens.eq(self.padding_idx)  # B, T
        x = self.embed_scale * self.embed_tokens(tokens)
        x.requires_grad = True
        
        if getattr(self.args, 'token_dropout', False):
            x.masked_fill_((tokens == self.mask_idx).unsqueeze(-1), 0.0)
            mask_ratio_train = 0.15 * 0.8
            src_lengths = (~padding_mask).sum(-1)
            mask_ratio_observed = (tokens == self.mask_idx).sum(-1).float() / src_lengths
            x = x * (1 - mask_ratio_train) / ((1 - mask_ratio_observed)[:, None, None]).type_as(x)

        x = x + self.embed_positions(tokens)

        if self.model_version == 'ESM-1b':
            x = self.emb_layer_norm_before(x)
            if padding_mask is not None:
                x = x * (1 - padding_mask.unsqueeze(-1).type_as(x))

        repr_layers = set(repr_layers)
        hidden_representations = {}
        if 0 in repr_layers:
            hidden_representations[0] = x

        # (B, T, E) => (T, B, E)
        x = x.transpose(0, 1)

        if not padding_mask.any():
            padding_mask = None

        for layer_idx, layer in enumerate(self.layers):
            # if use_checkpoint:
            #     x, _ = checkpoint(layer, x, None, padding_mask)
            # else:
            x, _ = layer(x, self_attn_padding_mask=padding_mask)
                
            if (layer_idx + 1) in repr_layers:
                hidden_representations[layer_idx + 1] = x.transpose(0, 1)

        if self.model_version == 'ESM-1b':
            x = self.emb_layer_norm_after(x)
            x = x.transpose(0, 1)# (T, B, E) => (B, T, E)

            # last hidden representation should have layer norm applied
            if (layer_idx + 1) in repr_layers:
                hidden_representations[layer_idx + 1] = x
            '''x3 = x * (1 - padding_mask.unsqueeze(-1).type_as(x))
            print(x3.size())'''

        result = {"representations": hidden_representations}
      
        return result

    def predict_contacts(self, tokens):
        return self(tokens, return_contacts=True)["contacts"]

    @property
    def num_layers(self):
        return self.args.layers


class MSATransformer(nn.Module):

    @classmethod
    def add_args(cls, parser):
        parser.add_argument(
            "--num_layers", default=12, type=int, metavar="N", help="number of layers"
        )
        parser.add_argument(
            "--embed_dim", default=768, type=int, metavar="N", help="embedding dimension"
        )
        parser.add_argument(
            "--logit_bias", action="store_true", help="whether to apply bias to logits"
        )
        parser.add_argument(
            "--ffn_embed_dim",
            default=3072,
            type=int,
            metavar="N",
            help="embedding dimension for FFN",
        )
        parser.add_argument(
            "--attention_heads",
            default=12,
            type=int,
            metavar="N",
            help="number of attention heads",
        )
        parser.add_argument(
            "--dropout",
            default=0.1,
            type=float,
            help="Dropout to apply."
        )
        parser.add_argument(
            "--attention_dropout",
            default=0.1,
            type=float,
            help="Dropout to apply."
        )
        parser.add_argument(
            "--activation_dropout",
            default=0.1,
            type=float,
            help="Dropout to apply."
        )
        parser.add_argument(
            "--max_tokens_per_msa",
            default=2 ** 14,
            type=int,
            help=(
                "Used during inference to batch attention computations in a single "
                "forward pass. This allows increased input sizes with less memory."
            ),
        )

    def __init__(self, args, alphabet):
        super().__init__()
        self.args = args
        self.alphabet_size = len(alphabet)
        self.padding_idx = alphabet.padding_idx
        self.mask_idx = alphabet.mask_idx
        self.cls_idx = alphabet.cls_idx
        self.eos_idx = alphabet.eos_idx
        self.prepend_bos = alphabet.prepend_bos
        self.append_eos = alphabet.append_eos

        self.embed_tokens = nn.Embedding(
            self.alphabet_size, self.args.embed_dim, padding_idx=self.padding_idx
        )

        if getattr(args, "embed_positions_msa", False):
            self.msa_position_embedding = nn.Parameter(
                0.01 * torch.randn(1, 1024, 1, 1),
                requires_grad=True,
            )
        else:
            self.register_parameter("msa_position_embedding", None)

        self.dropout_module = nn.Dropout(self.args.dropout)
        self.layers = nn.ModuleList(
            [
                AxialTransformerLayer(
                    self.args.embed_dim,
                    self.args.ffn_embed_dim,
                    self.args.attention_heads,
                    self.args.dropout,
                    self.args.attention_dropout,
                    self.args.activation_dropout,
                    getattr(self.args, "max_tokens_per_msa", self.args.max_tokens),
                )
                for _ in range(self.args.layers)
            ]
        )

        self.contact_head = ContactPredictionHead(
            self.args.layers * self.args.attention_heads,
            self.prepend_bos,
            self.append_eos,
            eos_idx=self.eos_idx,
        )
        self.embed_positions = LearnedPositionalEmbedding(
            self.args.max_positions, self.args.embed_dim, self.padding_idx,
        )
        self.emb_layer_norm_before = ESM1bLayerNorm(self.args.embed_dim)
        self.emb_layer_norm_after = ESM1bLayerNorm(self.args.embed_dim)
        self.lm_head = RobertaLMHead(
            embed_dim=self.args.embed_dim,
            output_dim=self.alphabet_size,
            weight=self.embed_tokens.weight
        )

    def forward(
        self,
        tokens,
        repr_layers=[],
        need_head_weights=False,
        return_contacts=False
    ):
        if return_contacts:
            need_head_weights = True

        assert tokens.ndim == 3
        batch_size, num_alignments, seqlen = tokens.size()
        padding_mask = tokens.eq(self.padding_idx)  # B, R, C
        if not padding_mask.any():
            padding_mask = None

        x = self.embed_tokens(tokens)
        x += self.embed_positions(
            tokens.view(batch_size * num_alignments, seqlen)
        ).view(x.size())
        if self.msa_position_embedding is not None:
            if x.size(1) > 1024:
                raise RuntimeError(
                    "Using model with MSA position embedding trained on maximum MSA "
                    f"depth of 1024, but received {x.size(1)} alignments."
                )
            x += self.msa_position_embedding[:, :num_alignments]

        x = self.emb_layer_norm_before(x)

        x = self.dropout_module(x)

        if padding_mask is not None:
            x = x * (1 - padding_mask.unsqueeze(-1).type_as(x))

        repr_layers = set(repr_layers)
        hidden_representations = {}
        if 0 in repr_layers:
            hidden_representations[0] = x

        if need_head_weights:
            row_attn_weights = []
            col_attn_weights = []

        # B x R x C x D -> R x C x B x D
        x = x.permute(1, 2, 0, 3)

        for layer_idx, layer in enumerate(self.layers):
            x = layer(
                x,
                self_attn_padding_mask=padding_mask,
                need_head_weights=need_head_weights,
            )
            if need_head_weights:
                x, col_attn, row_attn = x
                # H x C x B x R x R -> B x H x C x R x R
                col_attn_weights.append(col_attn.permute(2, 0, 1, 3, 4))
                # H x B x C x C -> B x H x C x C
                row_attn_weights.append(row_attn.permute(1, 0, 2, 3))
            if (layer_idx + 1) in repr_layers:
                hidden_representations[layer_idx + 1] = x.permute(2, 0, 1, 3)

        x = self.emb_layer_norm_after(x)
        x = x.permute(2, 0, 1, 3)  # R x C x B x D -> B x R x C x D

        # last hidden representation should have layer norm applied
        if (layer_idx + 1) in repr_layers:
            hidden_representations[layer_idx + 1] = x
        x = self.lm_head(x)

        result = {"logits": x, "representations": hidden_representations}
        if need_head_weights:
            # col_attentions: B x L x H x C x R x R
            col_attentions = torch.stack(col_attn_weights, 1)
            # row_attentions: B x L x H x C x C
            row_attentions = torch.stack(row_attn_weights, 1)
            result["col_attentions"] = col_attentions
            result["row_attentions"] = row_attentions
            if return_contacts:
                contacts = self.contact_head(
                    tokens, row_attentions
                )
                result["contacts"] = contacts

        return result

    def predict_contacts(self, tokens):
        return self(tokens, return_contacts=True)["contacts"]

    @property
    def num_layers(self):
        return self.args.layers

    def max_tokens_per_msa_(self, value: int) -> None:
        """ The MSA Transformer automatically batches attention computations when
        gradients are disabled to allow you to pass in larger MSAs at test time than
        you can fit in GPU memory. By default this occurs when more than 2^14 tokens
        are passed in the input MSA. You can set this value to infinity to disable
        this behavior.
        """
        for module in self.modules():
            if isinstance(module, (RowSelfAttention, ColumnSelfAttention)):
                module.max_tokens_per_msa = value
